from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    from typing import Optional
else:
    Optional, = jac_import('typing', 'py', items={'Optional': None})
if typing.TYPE_CHECKING:
    from jivas.agent.memory.memory_frame_walker import memory_frame_walker
else:
    memory_frame_walker, = jac_import('jivas.agent.memory.memory_frame_walker', items={'memory_frame_walker': None})
if typing.TYPE_CHECKING:
    from jivas.agent.memory.memory import Memory
else:
    Memory, = jac_import('jivas.agent.memory.memory', items={'Memory': None})

class update_frame(memory_frame_walker, Walker):
    session_id: str = field('')
    label: str = field('')
    user_name: str = field('')

    class __specs__(Obj):
        private: static[bool] = False

    @with_entry
    def on_memory(self, here: Memory) -> None:
        if not self.session_id:
            Jac.get_context().status = 400
            Jac.report('missing session_id')
            return
        if not self.label or self.user_name:
            Jac.get_context().status = 400
            Jac.report('nothing supplied to update frame')
            return
        frame_node = here.get_frame(agent_id=self.agent_id, session_id=self.session_id)
        if not frame_node:
            Jac.get_context().status = 500
            Jac.report('unable to update frame node')
            return
        frame_node.set_label(self.label)
        frame_node.set_user_name(self.user_name)
        frame_node.update()
        Jac.report(frame_node.export())


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    from jivas.agent.core.agent_graph_walker import agent_graph_walker
else:
    agent_graph_walker, = jac_import('jivas.agent.core.agent_graph_walker', items={'agent_graph_walker': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.app import App
else:
    App, = jac_import('jivas.agent.core.app', items={'App': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.agents import Agents
else:
    Agents, = jac_import('jivas.agent.core.agents', items={'Agents': None})

class list_agents(agent_graph_walker, Walker):

    class __specs__(Obj):
        private: static[bool] = False
        excluded: static[list] = JacList(['agent_id'])

    @with_entry
    def on_agents(self, here: Agents) -> None:
        if (agents := here.get_all()):
            for agent in agents:
                Jac.report(agent.export())


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    import io
else:
    io, = jac_import('io', 'py')
if typing.TYPE_CHECKING:
    import zipfile
else:
    zipfile, = jac_import('zipfile', 'py')
if typing.TYPE_CHECKING:
    from jivas.agent.core.agent_graph_walker import agent_graph_walker
else:
    agent_graph_walker, = jac_import('jivas.agent.core.agent_graph_walker', items={'agent_graph_walker': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.agent import Agent
else:
    Agent, = jac_import('jivas.agent.core.agent', items={'Agent': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.action import Action
else:
    Action, = jac_import('jivas.agent.action.action', items={'Action': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.actions import Actions
else:
    Actions, = jac_import('jivas.agent.action.actions', items={'Actions': None})
if typing.TYPE_CHECKING:
    from jivas.agent.modules.agentlib.utils import Utils
else:
    Utils, = jac_import('jivas.agent.modules.agentlib.utils', 'py', items={'Utils': None})

class export_daf(agent_graph_walker, Walker):
    clean: bool = field(False)
    with_knowledge: bool = field(True)
    with_memory: bool = field(False)

    class __specs__(Obj):
        private: static[bool] = False

    @with_entry
    def on_agent(self, here: Agent) -> None:
        daf_contents = {}
        daf_name = here.meta.get('namespace', 'agent_daf')
        daf_info_yaml = Utils.yaml_dumps(here.get_daf_info())
        daf_descriptor_yaml = here.get_descriptor(as_yaml=True, clean=self.clean)
        if daf_info_yaml and daf_descriptor_yaml:
            daf_contents = {'info.yaml': daf_info_yaml, 'descriptor.yaml': daf_descriptor_yaml}
        else:
            Jac.get_context().status = 503
            Jac.report('no valid daf info or descriptor generated, unable to complete export')
            return self.disengage()
        if self.with_memory:
            daf_memory_data = here.get_memory().export_memory()
            if (daf_memory_yaml := Utils.yaml_dumps(daf_memory_data)):
                daf_contents['memory.yaml'] = daf_memory_yaml
            else:
                self.logger.error('Unable to export memory. It may be blank or there may be a YAML conversion issue.')
        if self.with_knowledge:
            daf_knowledge = JacList([])
            try:
                if (vector_store_action := here.get_action(action_label=here.get_vector_store_action())):
                    if (daf_knowledge_yaml := vector_store_action.export_knodes()):
                        daf_contents['knowledge.yaml'] = daf_knowledge_yaml
                    else:
                        self.logger.error('Unable to export knowledge. It may be blank or there may be a YAML conversion issue.')
            except Exception as e:
                Jac.get_context().status = 503
                error_message = f'Unable to export knowledge : {e}'
                self.logger.error(error_message)
                Jac.report(error_message)
                return self.disengage()
        buffer = io.BytesIO()
        with zipfile.ZipFile(buffer, 'w', zipfile.ZIP_DEFLATED) as zipf:
            for fname, content in daf_contents.items():
                zipf.writestr(fname, content)
        daf_bytes = buffer.getvalue()
        daf_output_filename = f"dafs/{daf_name.replace('/', '_')}.daf.zip"
        if here.save_file(daf_output_filename, daf_bytes):
            Jac.get_context().status = 200
            Jac.report(here.get_file_url(daf_output_filename))
        else:
            Jac.get_context().status = 503
            error_message = 'There was a problem generating the DAF'
            self.logger.error(error_message)
            Jac.report(error_message)


from __future__ import annotations
from jaclang import *
import typing
if typing.TYPE_CHECKING:
    from jivas.agent.modules.agentlib.utils import Utils
else:
    Utils, = jac_import('jivas.agent.modules.agentlib.utils', 'py', items={'Utils': None})

class Data(Obj):
    label: str = field('')
    meta: dict = field(gen=lambda: {})
    content: any = field('')

    def load(self, data: dict) -> None:
        if data:
            for attr in data.keys():
                if hasattr(self, attr):
                    setattr(self, attr, data[attr])

    def export(self, ignore_keys: list=JacList(['__jac__'])) -> None:
        node_export = Utils.export_to_dict(self, ignore_keys)
        return node_export


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    import os
else:
    os, = jac_import('os', 'py')
if typing.TYPE_CHECKING:
    import yaml
else:
    yaml, = jac_import('yaml', 'py')
if typing.TYPE_CHECKING:
    import logging
else:
    logging, = jac_import('logging', 'py')
if typing.TYPE_CHECKING:
    import traceback
else:
    traceback, = jac_import('traceback', 'py')
if typing.TYPE_CHECKING:
    from logging import Logger
else:
    Logger, = jac_import('logging', 'py', items={'Logger': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.agent import Agent
else:
    Agent, = jac_import('jivas.agent.core.agent', items={'Agent': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.action import Action
else:
    Action, = jac_import('jivas.agent.action.action', items={'Action': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.actions import Actions
else:
    Actions, = jac_import('jivas.agent.action.actions', items={'Actions': None})
if typing.TYPE_CHECKING:
    from interact_graph_walker import interact_graph_walker
else:
    interact_graph_walker, = jac_import('interact_graph_walker', items={'interact_graph_walker': None})

class get_action_app(interact_graph_walker, Walker):
    agent_id: str = field('')
    action: str = field('')
    action_app: str = field('')
    logger: static[Logger] = logging.getLogger(__name__)

    class __specs__(Obj):
        private: static[bool] = False
        excluded: static[list] = JacList(['action_app'])

    @with_entry
    def on_agent(self, here: Agent) -> None:
        self.visit(here.refs().filter(Actions, None))

    @with_entry
    def on_actions(self, here: Actions) -> None:
        if not self.visit(here.refs().filter(Action, None).filter(None, lambda item: item.label == self.action)):
            message = f'Action not found: {self.action}'
            self.logger.error(message)
            Jac.get_context().status = 404
            Jac.report(message)

    @with_entry
    def on_action(self, here: Action) -> None:
        action_path = f"{here._package.get('config', {}).get('path', '')}/app/app.py"
        try:
            if os.path.exists(action_path):
                with open(action_path, 'r') as file:
                    self.action_app = file.read()
            else:
                message = f'Action app file not found: {action_path}'
                self.logger.warning(message)
                Jac.get_context().status = 404
                Jac.report(message)
                return
        except Exception as e:
            message = f'Error reading action app file: {str(e)}'
            self.logger.error(message)
            self.logger.error(traceback.format_exc())
            Jac.get_context().status = 500
            Jac.report(message)
            return
        if self.reporting:
            Jac.report(self.action_app)


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    from jivas.agent.core.graph_node import GraphNode
else:
    GraphNode, = jac_import('jivas.agent.core.graph_node', items={'GraphNode': None})

class Collection(GraphNode, Node):
    name: str = field('')
    data: dict = field(gen=lambda: {})

    def get_name(self) -> str:
        return self.name

    def set_name(self, name: str) -> None:
        self.name = name

    def get_data_item(self, label: str) -> any:
        return self.data.get(label, None)

    def set_data_item(self, label: str, value: any) -> None:
        self.data[label] = value

    def delete(self) -> list:
        return self.spawn(_purge_collection()).removed

class _purge_collection(Walker):
    removed: list = field(gen=lambda: JacList([]))

    class __specs__(Obj):
        private: static[bool] = True

    @with_entry
    def on_collection(self, here: Collection) -> None:
        if not self.visit(here.refs()):
            self.removed.append(here)
            Jac.destroy(here)

    @with_entry
    def on_collection_node(self, here: Node) -> None:
        self.visit(here.refs())
        self.removed.append(here)
        Jac.destroy(here)


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    from jivas.agent.core import init_agents, import_agent, export_descriptor, export_daf, get_agent, update_agent, list_agents, delete_agent, healthcheck
else:
    init_agents, import_agent, export_descriptor, export_daf, get_agent, update_agent, list_agents, delete_agent, healthcheck = jac_import('jivas.agent.core', items={'init_agents': None, 'import_agent': None, 'export_descriptor': None, 'export_daf': None, 'get_agent': None, 'update_agent': None, 'list_agents': None, 'delete_agent': None, 'healthcheck': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action import interact, pulse, list_actions, get_action, update_action, install_action, uninstall_action, invoke_stt_action, get_action_app
else:
    interact, pulse, list_actions, get_action, update_action, install_action, uninstall_action, invoke_stt_action, get_action_app = jac_import('jivas.agent.action', items={'interact': None, 'pulse': None, 'list_actions': None, 'get_action': None, 'update_action': None, 'install_action': None, 'uninstall_action': None, 'invoke_stt_action': None, 'get_action_app': None})
if typing.TYPE_CHECKING:
    from jivas.agent.memory import add_frame, get_frames, update_frame, delete_frame, get_interactions, update_interaction
else:
    add_frame, get_frames, update_frame, delete_frame, get_interactions, update_interaction = jac_import('jivas.agent.memory', items={'add_frame': None, 'get_frames': None, 'update_frame': None, 'delete_frame': None, 'get_interactions': None, 'update_interaction': None})
if typing.TYPE_CHECKING:
    from jivas.agent.analytics import get_channels_by_date, get_users_by_date, get_interactions_by_date, get_interaction_logs
else:
    get_channels_by_date, get_users_by_date, get_interactions_by_date, get_interaction_logs = jac_import('jivas.agent.analytics', items={'get_channels_by_date': None, 'get_users_by_date': None, 'get_interactions_by_date': None, 'get_interaction_logs': None})


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    import io
else:
    io, = jac_import('io', 'py')
if typing.TYPE_CHECKING:
    import os
else:
    os, = jac_import('os', 'py')
if typing.TYPE_CHECKING:
    import json
else:
    json, = jac_import('json', 'py')
if typing.TYPE_CHECKING:
    import yaml
else:
    yaml, = jac_import('yaml', 'py')
if typing.TYPE_CHECKING:
    import requests
else:
    requests, = jac_import('requests', 'py')
if typing.TYPE_CHECKING:
    import tarfile
else:
    tarfile, = jac_import('tarfile', 'py')
if typing.TYPE_CHECKING:
    import logging
else:
    logging, = jac_import('logging', 'py')
if typing.TYPE_CHECKING:
    import traceback
else:
    traceback, = jac_import('traceback', 'py')
if typing.TYPE_CHECKING:
    from typing import Union
else:
    Union, = jac_import('typing', 'py', items={'Union': None})
if typing.TYPE_CHECKING:
    from logging import Logger
else:
    Logger, = jac_import('logging', 'py', items={'Logger': None})
if typing.TYPE_CHECKING:
    from jivas.agent.modules.agentlib.utils import Utils
else:
    Utils, = jac_import('jivas.agent.modules.agentlib.utils', 'py', items={'Utils': None})
if typing.TYPE_CHECKING:
    from jvcli.api import RegistryAPI
else:
    RegistryAPI, = jac_import('jvcli.api', 'py', items={'RegistryAPI': None})
if typing.TYPE_CHECKING:
    from jvcli.utils import is_version_compatible
else:
    is_version_compatible, = jac_import('jvcli.utils', 'py', items={'is_version_compatible': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.app import App
else:
    App, = jac_import('jivas.agent.core.app', items={'App': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.agent import Agent
else:
    Agent, = jac_import('jivas.agent.core.agent', items={'Agent': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.agents import Agents
else:
    Agents, = jac_import('jivas.agent.core.agents', items={'Agents': None})
if typing.TYPE_CHECKING:
    from jivas.agent.memory.memory import Memory
else:
    Memory, = jac_import('jivas.agent.memory.memory', items={'Memory': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.action import Action
else:
    Action, = jac_import('jivas.agent.action.action', items={'Action': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.actions import Actions
else:
    Actions, = jac_import('jivas.agent.action.actions', items={'Actions': None})

class import_agent(Walker):
    logger: static[Logger] = logging.getLogger(__name__)
    descriptor: str = field('')
    daf_name: str = field('')
    daf_version: str = field('')
    jpr_api_key: str = field('')
    reporting: bool = field(True)

    class __specs__(Obj):
        private: static[bool] = False

    @with_entry
    def on_root(self, here: Root) -> None:
        if not self.visit(here.refs().filter(App, None)):
            self.logger.info('App node created')
            app_node = here.connect(App())
            self.visit(app_node)

    @with_entry
    def on_app(self, here: App) -> None:
        if not self.visit(here.refs().filter(Agents, None)):
            self.logger.info('Agents node created')
            agents_node = here.connect(Agents())
            self.visit(agents_node)

    @with_entry
    def on_agents(self, here: Agents) -> None:
        if self.descriptor:
            if (agent_node := self.import_from_descriptor(here, self.descriptor)):
                if self.reporting:
                    Jac.report(agent_node.export())
        elif self.daf_name:
            if (agent_node := self.import_from_daf(here, self.daf_name, self.daf_version, self.jpr_api_key)):
                if self.reporting:
                    Jac.report(agent_node.export())

    def get_daf_root(self) -> str:
        daf_root_path = os.environ.get('JIVAS_DAF_ROOT_PATH', 'daf')
        if not os.path.exists(daf_root_path):
            os.makedirs(daf_root_path)
        return daf_root_path

    def import_from_descriptor(self, agents_node: Agents, descriptor: Union[str, dict]) -> Agent:
        agent_data = {}
        agent_node = None
        if not descriptor:
            return None
        if isinstance(descriptor, str):
            try:
                agent_data = json.loads(descriptor)
            except json.JSONDecodeError:
                pass
            try:
                agent_data = yaml.safe_load(descriptor)
            except yaml.YAMLError:
                pass
        else:
            agent_data = descriptor
        if not agent_data:
            self.logger.error('no agent data available for import')
            return None
        if agent_data.get('id', None):
            try:
                agent_node = jobj(id=agent_data['id'])
            except Exception as e:
                self.logger.error(f'an exception occurred, {traceback.format_exc()}')
                return None
        else:
            agent_name = agent_data.get('name')
            if (_node := agents_node.get_by_name(agent_name)):
                agent_node = _node
            elif (agent_node := Agent(name=agent_name, description=agent_data.get('description', ''))):
                if not agents_node.get_by_id(agent_node.id):
                    agents_node.connect(agent_node)
                self.logger.info(f'agent created: {agent_node.name}')
            else:
                self.logger.error(f'unable to create agent: {agent_name}')
        if agent_node:
            descriptor_root = Utils.get_descriptor_root()
            agent_node.descriptor = f'{descriptor_root}/{agent_node.id}.yaml'
            agent_node.get_memory()
            agent_node.get_actions()
            if (agent_node := agent_node.update(data=agent_data, with_actions=True)):
                return agent_node
        return None

    def import_from_daf(self, agents_node: Agents, daf_name: str, version: Optional[str]='', jpr_api_key: Optional[str]='') -> Agent:
        agent_node = None
        safe_to_import = False
        daf_info = {}
        package_root = self.get_daf_root()
        namespace, package_name = daf_name.split('/')
        if (package_path := Utils.find_package_folder(package_root, daf_name)):
            info_yaml_path = os.path.join(package_path, 'info.yaml')
            if os.path.exists(info_yaml_path):
                with open(info_yaml_path, 'r') as file:
                    try:
                        daf_info = yaml.safe_load(file)
                        package_version = daf_info.get('package', {}).get('version', None)
                        if version:
                            if is_version_compatible(version, package_version):
                                safe_to_import = True
                        else:
                            safe_to_import = True
                    except yaml.YAMLError as e:
                        self.logger.error(f'an exception occurred, {traceback.format_exc()}')
        if not safe_to_import:
            if (package_data := RegistryAPI.download_package(daf_name, version, api_key=jpr_api_key)):
                try:
                    package_file = requests.get(package_data['file'])
                    target_dir = os.path.join(package_root, daf_name)
                    os.makedirs(target_dir, exist_ok=True)
                    with tarfile.open(fileobj=io.BytesIO(package_file.content), mode='r:gz') as tar_file:
                        tar_file.extractall(target_dir)
                    safe_to_import = True
                except Exception as e:
                    self.logger.error(f'an exception occurred, {traceback.format_exc()}')
        if safe_to_import and (package_path := Utils.find_package_folder(package_root, daf_name)):
            info_yaml_path = os.path.join(package_path, 'info.yaml')
            descriptor_yaml_path = os.path.join(package_path, 'descriptor.yaml')
            memory_yaml_path = os.path.join(package_path, 'memory.yaml')
            knowledge_yaml_path = os.path.join(package_path, 'knowledge.yaml')
            if os.path.exists(info_yaml_path):
                with open(info_yaml_path, 'r') as file:
                    try:
                        daf_info = yaml.safe_load(file)
                    except Exception as e:
                        self.logger.error(f'an exception occurred, {traceback.format_exc()}')
                        return None
            if os.path.exists(descriptor_yaml_path):
                with open(descriptor_yaml_path, 'r') as file:
                    try:
                        descriptor_data = yaml.safe_load(file)
                        if jpr_api_key:
                            descriptor_data['jpr_api_key'] = jpr_api_key
                        descriptor_data['meta'] = {'namespace': daf_info.get('package', {}).get('name'), 'version': daf_info.get('package', {}).get('version'), 'author': daf_info.get('package', {}).get('author'), 'dependencies': daf_info.get('package', {}).get('dependencies')}
                        agent_node = self.import_from_descriptor(agents_node, descriptor_data)
                    except Exception as e:
                        self.logger.error(f'an exception occurred, {traceback.format_exc()}')
                        return None
            if agent_node:
                if agent_node.healthcheck_status != 200:
                    self.logger.error(f'{agent_node.name} failed healthcheck. Any memory or knowledge queued for import will be deferred.')
                    return agent_node
            if agent_node and os.path.exists(memory_yaml_path):
                with open(memory_yaml_path, 'r') as file:
                    try:
                        _info = yaml.safe_load(file)
                        agent_node.get_memory().import_memory(data=_info)
                    except Exception as e:
                        self.logger.error(f'an exception occurred, {traceback.format_exc()}')
            if agent_node and os.path.exists(knowledge_yaml_path):
                with open(knowledge_yaml_path, 'r') as file:
                    try:
                        knode_yaml = file.read()
                        knode_yaml_data = yaml.safe_load(knode_yaml)
                        if (vector_store_action := agent_node.get_vector_store_action()):
                            if 'vec' in knode_yaml_data[0]:
                                vector_store_action.add_embeddings(knode_yaml)
                            else:
                                vector_store_action.import_knodes(knode_yaml)
                    except Exception as e:
                        self.logger.error(f'an exception occurred while importing agent knowledge, {traceback.format_exc()}')
        return agent_node


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    import logging
else:
    logging, = jac_import('logging', 'py')
if typing.TYPE_CHECKING:
    from logging import Logger
else:
    Logger, = jac_import('logging', 'py', items={'Logger': None})
if typing.TYPE_CHECKING:
    from jivas.agent.memory.interaction import Interaction
else:
    Interaction, = jac_import('jivas.agent.memory.interaction', items={'Interaction': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.graph_walker import graph_walker
else:
    graph_walker, = jac_import('jivas.agent.core.graph_walker', items={'graph_walker': None})
if typing.TYPE_CHECKING:
    from jac_cloud.core.architype import NodeAnchor
else:
    NodeAnchor, = jac_import('jac_cloud.core.architype', 'py', items={'NodeAnchor': None})

class update_interaction(graph_walker, Walker):
    logger: static[Logger] = logging.getLogger(__name__)
    interaction_data: dict = field(gen=lambda: {})

    class __specs__(Obj):
        auth: static[bool] = True
        private: static[bool] = False

    @with_entry
    def on_interaction(self, here: Interaction) -> None:
        if not self.interaction_data:
            self.logger.error('no interaction data')
            return
        here.update(data=self.interaction_data)
        here.close()


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    import logging
else:
    logging, = jac_import('logging', 'py')
if typing.TYPE_CHECKING:
    import traceback
else:
    traceback, = jac_import('traceback', 'py')
if typing.TYPE_CHECKING:
    from logging import Logger
else:
    Logger, = jac_import('logging', 'py', items={'Logger': None})
if typing.TYPE_CHECKING:
    from jivas.agent.modules.agentlib.utils import Utils
else:
    Utils, = jac_import('jivas.agent.modules.agentlib.utils', 'py', items={'Utils': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.app import App
else:
    App, = jac_import('jivas.agent.core.app', items={'App': None})

class graph_walker(Walker):
    reporting: bool = field(True)
    logger: static[Logger] = logging.getLogger(__name__)

    class __specs__(Obj):
        private: static[bool] = True

    @with_entry
    def on_root(self, here: Root) -> None:
        if not self.visit(here.refs().filter(App, None)):
            self.logger.error('App graph not initialized. Import an agent and try again.')

    def export(self, ignore_keys: list=JacList(['__jac__'])) -> None:
        node_export = Utils.export_to_dict(self, ignore_keys)
        return node_export

    def update(self, data: dict={}) -> graph_walker:
        if data:
            for attr in data.keys():
                if hasattr(self, attr):
                    self.attr = data[attr]
        self.post_update()
        return self

    def post_update(self) -> None:
        pass


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    from jivas.agent.core.agent import Agent
else:
    Agent, = jac_import('jivas.agent.core.agent', items={'Agent': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.action import Action
else:
    Action, = jac_import('jivas.agent.action.action', items={'Action': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.actions import Actions
else:
    Actions, = jac_import('jivas.agent.action.actions', items={'Actions': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.interact_graph_walker import interact_graph_walker
else:
    interact_graph_walker, = jac_import('jivas.agent.action.interact_graph_walker', items={'interact_graph_walker': None})
if typing.TYPE_CHECKING:
    import os
else:
    os, = jac_import('os', 'py')
if typing.TYPE_CHECKING:
    import uuid
else:
    uuid, = jac_import('uuid', 'py')
if typing.TYPE_CHECKING:
    import base64
else:
    base64, = jac_import('base64', 'py')

class invoke_stt_action(interact_graph_walker, Walker):
    files: list[dict] = field(gen=lambda: JacList([]))

    @with_entry
    def on_agent(self, here: Agent) -> None:
        action_node = here.get_stt_action()
        if not self.files:
            self.response = {'success': False, 'message': 'No audio file provided'}
        else:
            file = self.files[0]
            if 'audio' in file['type'] or file['type'] == 'application/octet-stream':
                result = action_node.invoke_file(audio_content=file['content'], audio_type=file['type'])
                self.response = {'success': True, 'duration': result['duration'], 'transcript': result['transcript']}
            else:
                self.response = {'success': False, 'message': 'Invalid file type, must be audio'}


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    import os
else:
    os, = jac_import('os', 'py')
if typing.TYPE_CHECKING:
    import yaml
else:
    yaml, = jac_import('yaml', 'py')
if typing.TYPE_CHECKING:
    import io
else:
    io, = jac_import('io', 'py')
if typing.TYPE_CHECKING:
    import logging
else:
    logging, = jac_import('logging', 'py')
if typing.TYPE_CHECKING:
    import traceback
else:
    traceback, = jac_import('traceback', 'py')
if typing.TYPE_CHECKING:
    from logging import Logger
else:
    Logger, = jac_import('logging', 'py', items={'Logger': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.agent import Agent
else:
    Agent, = jac_import('jivas.agent.core.agent', items={'Agent': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.agents import Agents
else:
    Agents, = jac_import('jivas.agent.core.agents', items={'Agents': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.agent_graph_walker import agent_graph_walker
else:
    agent_graph_walker, = jac_import('jivas.agent.core.agent_graph_walker', items={'agent_graph_walker': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.import_agent import import_agent
else:
    import_agent, = jac_import('jivas.agent.core.import_agent', items={'import_agent': None})
if typing.TYPE_CHECKING:
    from jivas.agent.modules.agentlib.utils import Utils, jvdata_file_interface
else:
    Utils, jvdata_file_interface = jac_import('jivas.agent.modules.agentlib.utils', 'py', items={'Utils': None, 'jvdata_file_interface': None})

class init_agents(agent_graph_walker, Walker):
    reporting: bool = field(False)
    logger: static[Logger] = logging.getLogger(__name__)

    class __specs__(Obj):
        private: static[bool] = False
        excluded: static[list] = JacList(['agent_id'])

    @with_entry
    def on_agents(self, here: Agents) -> None:
        self.visit(here.refs().filter(Agent, None))

    @with_entry
    def on_agent(self, here: Agent) -> None:
        try:
            self.logger.info(f'initializing agent {here.name}')
            file_bytes = here.get_file(here.descriptor) or jvdata_file_interface.get_file(here.descriptor)
            if not file_bytes:
                self.logger.error(f'agent descriptor not found: {here.descriptor}')
                return
            file = io.BytesIO(file_bytes)
            descriptor = yaml.safe_load(file)
            agents_node = Utils.node_obj(here.refs(dir=EdgeDir.IN).filter(Agents, None))
            if descriptor:
                agents_node.spawn(import_agent(descriptor=descriptor, reporting=self.reporting))
        except Exception as e:
            self.logger.error(f'an exception occurred, {traceback.format_exc()}')


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    from jivas.agent.memory.memory_frame_walker import memory_frame_walker
else:
    memory_frame_walker, = jac_import('jivas.agent.memory.memory_frame_walker', items={'memory_frame_walker': None})
if typing.TYPE_CHECKING:
    from jivas.agent.memory.interaction import Interaction
else:
    Interaction, = jac_import('jivas.agent.memory.interaction', items={'Interaction': None})
if typing.TYPE_CHECKING:
    from jivas.agent.memory.advance import Advance
else:
    Advance, = jac_import('jivas.agent.memory.advance', items={'Advance': None})
if typing.TYPE_CHECKING:
    from jivas.agent.memory.memory import Memory
else:
    Memory, = jac_import('jivas.agent.memory.memory', items={'Memory': None})
if typing.TYPE_CHECKING:
    from jivas.agent.memory.frame import Frame
else:
    Frame, = jac_import('jivas.agent.memory.frame', items={'Frame': None})

class get_interactions(memory_frame_walker, Walker):
    session_id: str = field('')

    class __specs__(Obj):
        private: static[bool] = False

    @with_entry
    def on_memory(self, here: Memory) -> None:
        if self.session_id:
            self.visit(here.refs().filter(Frame, None).filter(None, lambda item: item.session_id == self.session_id))
        else:
            self.visit(here.refs().filter(Frame, None))

    @with_entry
    def on_frame(self, here: Frame) -> None:
        self.visit(here.refs(Advance).filter(Interaction, None))

    @with_entry
    def on_interaction(self, here: Interaction) -> None:
        Jac.report(here.export())
        self.visit(here.refs(Advance).filter(Interaction, None))


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    from jivas.agent.core.agent import Agent
else:
    Agent, = jac_import('jivas.agent.core.agent', items={'Agent': None})
if typing.TYPE_CHECKING:
    from jivas.agent.memory.memory import Memory
else:
    Memory, = jac_import('jivas.agent.memory.memory', items={'Memory': None})
if typing.TYPE_CHECKING:
    from jivas.agent.memory.frame import Frame
else:
    Frame, = jac_import('jivas.agent.memory.frame', items={'Frame': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.agent_graph_walker import agent_graph_walker
else:
    agent_graph_walker, = jac_import('jivas.agent.core.agent_graph_walker', items={'agent_graph_walker': None})

class memory_frame_walker(agent_graph_walker, Walker):
    session_id: str = field('')

    @with_entry
    def on_agent(self, here: Agent) -> None:
        self.visit(here.refs().filter(Memory, None))

    @with_entry
    def on_memory(self, here: Memory) -> None:
        self.visit(here.refs().filter(Frame, None).filter(None, lambda item: item.session_id == self.session_id))


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    from typing import Optional
else:
    Optional, = jac_import('typing', 'py', items={'Optional': None})
if typing.TYPE_CHECKING:
    from jivas.agent.memory.memory_frame_walker import memory_frame_walker
else:
    memory_frame_walker, = jac_import('jivas.agent.memory.memory_frame_walker', items={'memory_frame_walker': None})
if typing.TYPE_CHECKING:
    from jivas.agent.memory.memory import Memory
else:
    Memory, = jac_import('jivas.agent.memory.memory', items={'Memory': None})

class add_frame(memory_frame_walker, Walker):
    label: str = field('')
    user_name: str = field('')
    session_id: Optional[str] = field('')
    force_session: Optional[bool] = field(True)

    class __specs__(Obj):
        private: static[bool] = False

    @with_entry
    def on_memory(self, here: Memory) -> None:
        frame_node = here.get_frame(agent_id=self.agent_id, user_name=self.user_name, label=self.label, session_id=self.session_id, force_session=self.force_session)
        if not frame_node:
            Jac.get_context().status = 500
            Jac.report('unable to add frame node')
            return
        Jac.report(frame_node.export())


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    import os
else:
    os, = jac_import('os', 'py')
if typing.TYPE_CHECKING:
    import pytz
else:
    pytz, = jac_import('pytz', 'py')
if typing.TYPE_CHECKING:
    import json
else:
    json, = jac_import('json', 'py')
if typing.TYPE_CHECKING:
    import logging
else:
    logging, = jac_import('logging', 'py')
if typing.TYPE_CHECKING:
    import traceback
else:
    traceback, = jac_import('traceback', 'py')
if typing.TYPE_CHECKING:
    from typing import Optional
else:
    Optional, = jac_import('typing', 'py', items={'Optional': None})
if typing.TYPE_CHECKING:
    from logging import Logger
else:
    Logger, = jac_import('logging', 'py', items={'Logger': None})
if typing.TYPE_CHECKING:
    from jivas.agent.modules.agentlib.utils import Utils
else:
    Utils, = jac_import('jivas.agent.modules.agentlib.utils', 'py', items={'Utils': None})
if typing.TYPE_CHECKING:
    from datetime import datetime, timezone, timedelta
else:
    datetime, timezone, timedelta = jac_import('datetime', 'py', items={'datetime': None, 'timezone': None, 'timedelta': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.actions import Actions
else:
    Actions, = jac_import('jivas.agent.action.actions', items={'Actions': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.interact_action import InteractAction
else:
    InteractAction, = jac_import('jivas.agent.action.interact_action', items={'InteractAction': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.exit_interact_action import ExitInteractAction
else:
    ExitInteractAction, = jac_import('jivas.agent.action.exit_interact_action', items={'ExitInteractAction': None})
if typing.TYPE_CHECKING:
    from jivas.agent.memory.memory import Memory
else:
    Memory, = jac_import('jivas.agent.memory.memory', items={'Memory': None})
if typing.TYPE_CHECKING:
    from jivas.agent.memory.frame import Frame
else:
    Frame, = jac_import('jivas.agent.memory.frame', items={'Frame': None})
if typing.TYPE_CHECKING:
    from jivas.agent.memory.interaction import Interaction
else:
    Interaction, = jac_import('jivas.agent.memory.interaction', items={'Interaction': None})
if typing.TYPE_CHECKING:
    from jivas.agent.memory.interaction_response import InteractionResponse, InteractionMessage, SilentInteractionMessage
else:
    InteractionResponse, InteractionMessage, SilentInteractionMessage = jac_import('jivas.agent.memory.interaction_response', items={'InteractionResponse': None, 'InteractionMessage': None, 'SilentInteractionMessage': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.agent import Agent
else:
    Agent, = jac_import('jivas.agent.core.agent', items={'Agent': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.agents import Agents
else:
    Agents, = jac_import('jivas.agent.core.agents', items={'Agents': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.app import App
else:
    App, = jac_import('jivas.agent.core.app', items={'App': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action import interact_graph_walker
else:
    interact_graph_walker, = jac_import('jivas.agent.action', items={'interact_graph_walker': None})
if typing.TYPE_CHECKING:
    from jac_cloud.core.architype import NodeAnchor
else:
    NodeAnchor, = jac_import('jac_cloud.core.architype', 'py', items={'NodeAnchor': None})

class interact(interact_graph_walker, Walker):
    logger: static[Logger] = logging.getLogger(__name__)
    agent_id: str = field('')
    session_id: str = field('')
    utterance: str = field('')
    channel: str = field('default')
    data: list[dict] = field(gen=lambda: JacList([]))
    verbose: bool = field(False)
    tts: bool = field(False)
    streaming: bool = field(False)
    generator: any = field(None)
    response: dict = field(gen=lambda: {})
    message: InteractionMessage = field(None)
    execute: bool = field(False)
    context_data: dict = field(gen=lambda: {})
    frame_node: Frame = field(None)
    interaction_node: Interaction = field(None)
    agent_node: Agent = field(None)

    class __specs__(Obj):
        auth: static[bool] = True
        private: static[bool] = True
        excluded: static[list] = JacList(['response', 'message', 'execute', 'context_data', 'frame_node', 'interaction_node', 'agent_node'])

    @with_entry
    def on_root(self, here: Root) -> None:
        if self.init_interaction():
            self.visit(self.agent_node)
        else:
            return self.disengage()

    @with_entry
    def on_agent(self, here: Agent) -> None:
        self.visit(here.refs().filter(Actions, None))

    @with_entry
    def on_actions(self, here: Actions) -> None:
        last_interaction_node = self.frame_node.get_last_interaction(retraces=2)
        if (resume_action_node := self.frame_node.get_resume_action(last_interaction_node)):
            queued_interact_actions = here.queue_interact_actions(here.refs().filter(InteractAction, None).filter(None, lambda item: item.enabled == True))
            root_action = resume_action_node.get_root_action()
            trimmed_queue = self.trim_interact_actions(index=root_action, interact_actions=queued_interact_actions)
            if trimmed_queue:
                self.interaction_node.add_intent(root_action.label)
                self.interaction_node.context_data['resumed'] = True
                self.visit(trimmed_queue)
            else:
                self.visit(queued_interact_actions)
        else:
            queued_actions = here.queue_interact_actions(here.refs().filter(InteractAction, None).filter(None, lambda item: item.enabled == True))
            self.visit(queued_actions)

    @with_entry
    def on_action(self, here: InteractAction) -> None:
        result = {}
        intended = self.is_intended(here)
        has_access = self.has_access(here)
        has_function = self.has_function(here)
        if intended:
            if has_access:
                self.logger.debug(f'Touching: {here.label}')
                self.execute = here.touch(self)
                if self.execute:
                    self.logger.debug(f'Executing: {here.label}')
                    here.execute(self)
                    self.interaction_node.trail.append(here.label)
                    self.execute = False
            else:
                here.deny(self)

    @with_exit
    def on_exit(self, here) -> None:
        self.respond()

    def init_interaction(self) -> bool:
        try:
            self.agent_node = jobj(id=self.agent_id)
            if not self.agent_node:
                return False
        except Exception as e:
            Jac.get_context().status = 400
            self.logger.error('invalid agent id')
            return False
        if not self.agent_node.has_channel(self.channel):
            self.channel = 'default'
        self.frame_node = self.agent_node.get_memory().get_frame(agent_id=self.agent_node.id, session_id=self.session_id)
        if not self.frame_node:
            Jac.get_context().status = 500
            self.logger.error('unable to initiate a frame')
            return False
        if not self.is_valid_message_length(self.utterance, self.agent_node.message_limit):
            Jac.get_context().status = 400
            self.logger.warning('unable to process message; message length exceeds limit')
            return False
        if self.is_flood_active():
            Jac.get_context().status = 429
            self.logger.warning(f'flood control active on {self.frame_node.session_id}')
            return False
        add_interaction = False
        self.interaction_node = self.frame_node.get_last_interaction()
        if self.interaction_node is None:
            add_interaction = True
        elif self.interaction_node.is_closed():
            add_interaction = True
        if add_interaction:
            self.interaction_node = self.frame_node.add_interaction(utterance=self.utterance, channel=self.channel)
        if not self.interaction_node:
            return False
        if self.data and isinstance(self.data, list):
            required_fields = JacList(['label', 'meta', 'content'])
            if all(JacList([isinstance(item, dict) for item in self.data])):
                for item in self.data:
                    if all(JacList([field in item for field in required_fields])):
                        self.interaction_node.set_data_item(label=item['label'], meta=item['meta'], content=item['content'])
                    else:
                        Jac.get_context().status = 400
                        self.logger.error(f'malformed data item; requires {required_fields} in each dictionary')
                        return False
            else:
                Jac.get_context().status = 400
                self.logger.error("data must be a list of dictionaries with required fields: 'label', 'meta', 'content'")
                return False
        return True

    def is_valid_message_length(self, utterance: str, max_length: int) -> bool:
        chunks = Utils.chunk_long_message(message=utterance, max_length=max_length, chunk_length=max_length)
        if len(chunks) > 1:
            return False
        else:
            return True

    def is_flood_active(self) -> None:
        flood_control = self.agent_node.flood_control
        flood_block_time = self.agent_node.flood_block_time
        window_time = self.agent_node.window_time
        flood_threshold = self.agent_node.flood_threshold
        if flood_control:
            utc = pytz.UTC
            now = datetime.now(timezone.utc)
            flood = self.frame_node.variable_get(key='flood')
            if flood:
                expiration = flood['expiration']
                if isinstance(expiration, str):
                    expiration = datetime.strptime(expiration, '%Y-%m-%d %H:%M:%S.%f').replace(tzinfo=utc)
                elif isinstance(expiration, datetime) and expiration.tzinfo is None:
                    expiration = expiration.replace(tzinfo=utc)
                if expiration > now:
                    return True
            message_window = self.frame_node.variable_get(key='message_window')
            if message_window:
                for key in JacList(['start', 'end']):
                    value = message_window.get(key)
                    if isinstance(value, str):
                        message_window[key] = datetime.strptime(value, '%Y-%m-%d %H:%M:%S.%f').replace(tzinfo=utc)
                    elif isinstance(value, datetime):
                        if value.tzinfo is None:
                            message_window[key] = value.replace(tzinfo=utc)
                        else:
                            message_window[key] = value
                if message_window.get('end', datetime.min.replace(tzinfo=utc)) <= now:
                    message_window['start'] = now
                    message_window['end'] = now + timedelta(seconds=window_time)
                    message_window['count'] = 0
                else:
                    message_window['count'] = message_window.get('count', 0) + 1
                if message_window.get('count', 0) > flood_threshold:
                    flood = {'expiration': now + timedelta(seconds=flood_block_time)}
                    self.frame_node.variable_set(key='flood', value=flood)
                    return True
                self.frame_node.variable_set(key='message_window', value=message_window)
            else:
                message_window = {'start': now, 'end': now + timedelta(seconds=window_time), 'count': 1}
                self.frame_node.variable_set(key='message_window', value=message_window)
            return False
        else:
            return False

    def set_next_action(self, action_label: str, action_node: Optional[InteractAction]=None) -> None:
        next_action_node = None
        if action_label:
            next_action_node = self.agent_node.get_action(action_label=action_label)
        elif action_node:
            next_action_node = action_node
        if next_action_node:
            self.interaction_node.add_intent(next_action_node.label)
            self.prepend_interact_action(next_action_node)

    def set_resume_action(self, action_label: str, action_node: Optional[InteractAction]=None) -> None:
        resume_action_label = None
        if action_label:
            resume_action_label = action_label
        elif action_node:
            resume_action_label = action_node.label
        if resume_action_label:
            self.frame_node.set_resume_action(action_label=resume_action_label)

    def append_action(self, action_label: str, action_node: Optional[InteractAction]=None) -> None:
        append_action_node = None
        if action_label:
            append_action_node = self.agent_node.get_action(action_label=action_label)
        elif action_node:
            append_action_node = action_node
        if append_action_node:
            self.interaction_node.add_intent(append_action_node.label)
            self.append_interact_action(append_action_node)

    def dequeue_action(self, action_label: str, action_node: Optional[InteractAction]=None) -> None:
        if action_label:
            action_node = self.agent_node.get_action(action_label=action_label)
        if action_node:
            self.dequeue_interact_action(action_node)

    def append_interact_action(self, interact_action: InteractAction) -> None:
        if (exit_action := self.agent_node.get_action(action_label='ExitInteractAction')):
            exit_action_node_ref = None
            for i, node_ref in enumerate(self.__jac__.next):
                if f'{node_ref.id}' == f'{exit_action.__jac__.id}':
                    exit_action_node_ref = self.__jac__.next[i]
                    self.__jac__.next.pop(i)
                    break
            self.dequeue_interact_action(interact_action)
            self.visit(interact_action)
            self.__jac__.next.append(exit_action_node_ref)
            return self.__jac__.next
        else:
            self.visit(interact_action)
        return None

    def dequeue_interact_action(self, interact_action: InteractAction) -> None:
        if self.__jac__.next and interact_action:
            for i, node_ref in enumerate(self.__jac__.next):
                if f'{node_ref.id}' == f'{interact_action.__jac__.id}':
                    self.__jac__.next.pop(i)
                    break
            return self.__jac__.next
        return None

    def prepend_interact_action(self, interact_action: InteractAction) -> None:
        if self.dequeue_interact_action(interact_action):
            path = self.__jac__.next
            self.__jac__.next = JacList([])
            self.visit(interact_action)
            self.__jac__.next.extend(path)
            return self.__jac__.next
        return None

    def trim_interact_actions(self, index: InteractAction, interact_actions: list) -> list:
        trimmed = JacList([])
        found = False
        for action in interact_actions:
            if not found and f'{action.id}' == f'{index.id}':
                found = True
            if found:
                trimmed.append(action)
        return trimmed

    def has_access(self, action_node: Action) -> bool:
        access = True
        if action_node.get_type() in JacList(['AccessControlAction', 'ExitInteractAction']):
            return True
        if (access_control_action_node := self.agent_node.get_action(action_label='AccessControlAction')):
            access = access_control_action_node.has_action_access(session_id=self.frame_node.session_id, action_label=action_node.get_type(), channel=self.interaction_node.channel or 'default')
        return access

    def has_function(self, action_node: InteractAction) -> bool:
        if action_node.get_type() in JacList(['IntentInteractAction', 'FunctionInteractAction', 'ExitInteractAction']):
            return True
        if (function_interact_action_node := self.agent_node.get_action(action_label='FunctionInteractAction')):
            if function_interact_action_node.enabled and function_interact_action_node.strict:
                if action_node.label not in self.interaction_node.get_intents() + function_interact_action_node.exceptions:
                    return False
        return True

    def is_intended(self, action_node: InteractAction) -> bool:
        if action_node.get_type() in JacList(['IntentInteractAction', 'FunctionInteractAction', 'ExitInteractAction']):
            return True
        if (intent_interact_action_node := self.agent_node.get_action(action_label='IntentInteractAction')):
            if intent_interact_action_node.enabled and intent_interact_action_node.strict:
                if action_node.label not in self.interaction_node.get_intents() + intent_interact_action_node.exceptions:
                    return False
        return True

    def respond(self) -> None:
        if self.agent_node and self.interaction_node:
            self.interaction_node.close()
            interaction_data = {}
            has_response = self.interaction_node.has_response()
            if has_response:
                try:
                    self.message = self.interaction_node.get_message()
                    interaction_data = self.interaction_node.export()
                    if self.verbose:
                        interaction_data['frame'] = self.frame_node.export()
                        self.response = interaction_data
                    else:
                        self.response = {'response': self.interaction_node.get_response().export()}
                    if self.agent_node.is_logging():
                        self.log_interaction(data=interaction_data)
                except Exception as e:
                    self.logger.error(f'an exception occurred, {traceback.format_exc()}')
            if not has_response or not self.response:
                self.message = SilentInteractionMessage()
                self.response = InteractionResponse(session_id=self.frame_node.session_id, message=self.message.export()).export()
            self.frame_node.prune_interactions(frame_size=self.agent_node.frame_size)
        else:
            self.message = SilentInteractionMessage()
            self.response = InteractionResponse(message=self.message.export()).export()
        if self.tts:
            if (tts_action := self.agent_node.get_tts_action()):
                content = self.message.get_meta('phoneme_content') or self.message.get_content()
                audio = tts_action.invoke(text=content, as_url=True)
                if audio:
                    self.response['response'].update({'audio_url': audio})
        if self.reporting:
            Jac.report(self.response)

    def log_interaction(self, data: dict) -> None:
        collection = NodeAnchor.Collection.get_collection('interactions')
        collection.insert_one(json.loads(json.dumps(data)))


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    import logging
else:
    logging, = jac_import('logging', 'py')
if typing.TYPE_CHECKING:
    import traceback
else:
    traceback, = jac_import('traceback', 'py')
if typing.TYPE_CHECKING:
    from logging import Logger
else:
    Logger, = jac_import('logging', 'py', items={'Logger': None})
if typing.TYPE_CHECKING:
    from jivas.agent.modules.agentlib.utils import Utils
else:
    Utils, = jac_import('jivas.agent.modules.agentlib.utils', 'py', items={'Utils': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.graph_node import GraphNode
else:
    GraphNode, = jac_import('jivas.agent.core.graph_node', items={'GraphNode': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.purge import purge
else:
    purge, = jac_import('jivas.agent.core.purge', items={'purge': None})

class Agents(GraphNode, Node):
    logger: static[Logger] = logging.getLogger(__name__)

    def get_all(self) -> list:
        return self.refs()

    def get_by_name(self, name: str) -> None:
        return Utils.node_obj(self.refs().filter(None, lambda item: item.name == name))

    def get_by_id(self, id: str) -> None:
        return Utils.node_obj(self.refs().filter(None, lambda item: item.id == id))

    def delete(self, id: str) -> None:
        if (agent_node := self.get_by_id(id)):
            agent_node.get_actions().deregister_actions()
            agent_node.spawn(purge())
            return agent_node
        return None

    def delete_all(self) -> None:
        agent_nodes = JacList([])
        for agent_node in self.get_all():
            agent_nodes.append(self.delete(agent_node.id))
        return agent_nodes


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    import logging
else:
    logging, = jac_import('logging', 'py')
if typing.TYPE_CHECKING:
    import traceback
else:
    traceback, = jac_import('traceback', 'py')
if typing.TYPE_CHECKING:
    import os
else:
    os, = jac_import('os', 'py')
if typing.TYPE_CHECKING:
    from logging import Logger
else:
    Logger, = jac_import('logging', 'py', items={'Logger': None})
if typing.TYPE_CHECKING:
    from jivas.agent.modules.agentlib.utils import Utils
else:
    Utils, = jac_import('jivas.agent.modules.agentlib.utils', 'py', items={'Utils': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core import graph_node, purge
else:
    graph_node, purge = jac_import('jivas.agent.core', items={'graph_node': None, 'purge': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.graph_node import GraphNode
else:
    GraphNode, = jac_import('jivas.agent.core.graph_node', items={'GraphNode': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.action import Action
else:
    Action, = jac_import('jivas.agent.action.action', items={'Action': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.interact_action import InteractAction
else:
    InteractAction, = jac_import('jivas.agent.action.interact_action', items={'InteractAction': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.actions import Actions
else:
    Actions, = jac_import('jivas.agent.action.actions', items={'Actions': None})
if typing.TYPE_CHECKING:
    from jivas.agent.memory.memory import Memory
else:
    Memory, = jac_import('jivas.agent.memory.memory', items={'Memory': None})
if typing.TYPE_CHECKING:
    from typing import Union
else:
    Union, = jac_import('typing', 'py', items={'Union': None})
if typing.TYPE_CHECKING:
    from jvserve.lib.file_interface import file_interface
else:
    file_interface, = jac_import('jvserve.lib.file_interface', 'py', items={'file_interface': None})
if typing.TYPE_CHECKING:
    from jac_cloud.core.architype import NodeAnchor
else:
    NodeAnchor, = jac_import('jac_cloud.core.architype', 'py', items={'NodeAnchor': None})

class Agent(GraphNode, Node):
    logger: static[Logger] = logging.getLogger(__name__)
    published: bool = field(True)
    name: str = field('')
    description: str = field('')
    descriptor: str = field('')
    jpr_api_key: str = field('')
    agent_logging: bool = field(True)
    message_limit: int = field(1024)
    flood_control: bool = field(True)
    flood_block_time: int = field(300)
    window_time: int = field(20)
    flood_threshold: int = field(4)
    frame_size: int = field(10)
    tts_action: str = field('ElevenLabsTTSAction')
    stt_action: str = field('DeepgramSTTAction')
    vector_store_action: str = field('TypesenseVectorStoreAction')
    channels: list = field(gen=lambda: JacList(['default', 'whatsapp', 'facebook', 'slack', 'sms', 'email']))
    healthcheck_status: int = field(501)
    meta: dict = field(gen=lambda: {})

    def __post_init__(self) -> None:
        super().__post_init__()
        self.protected_attrs += JacList(['id', 'actions', 'descriptor', 'healthcheck_status'])

    def get_memory(self) -> Memory:
        memory_node = Utils.node_obj(self.refs().filter(Memory, None))
        if not memory_node:
            self.connect((memory_node := Memory()))
            self.logger.debug('memory node created')
        return memory_node

    def get_actions(self) -> Actions:
        actions_node = Utils.node_obj(self.refs().filter(Actions, None))
        if not actions_node:
            self.connect((actions_node := Actions()))
            self.logger.debug('actions node created')
        return actions_node

    def get_action(self, action_label: str='', action_type: str='', only_enabled: bool=True) -> Union[Action, None, list]:
        if (action_node := self.get_actions().get(action_label=action_label, action_type=action_type, only_enabled=only_enabled)):
            return action_node
        return None

    def get_tts_action(self) -> Union[Action, None]:
        if (tts_action_node := self.get_action(action_label=self.tts_action)):
            return tts_action_node
        return None

    def get_stt_action(self) -> Union[Action, None]:
        if (stt_action_node := self.get_action(action_label=self.stt_action)):
            return stt_action_node
        return None

    def get_vector_store_action(self) -> Union[Action, None]:
        if (vector_store_action_node := self.get_action(action_label=self.vector_store_action)):
            return vector_store_action_node
        return None

    def add_channel(self, channel: str) -> bool:
        if isinstance(channel, str) and channel.strip():
            channel = channel.strip().lower()
            if channel not in self.channels:
                self.channels.append(channel)
                return True
        return False

    def get_channels(self) -> list:
        return self.channels

    def remove_channel(self, channel: str) -> bool:
        if isinstance(channel, str) and channel.strip():
            channel = channel.strip().lower()
            if channel in self.channels:
                self.channels.remove(channel)
                return True
        return False

    def has_channel(self, channel: str) -> bool:
        if isinstance(channel, str) and channel.strip():
            return channel.strip().lower() in self.channels
        return False

    def validate_channels(self) -> list:
        valid_channels = JacList([])
        for channel in self.channels:
            if isinstance(channel, str) and channel.strip():
                valid_channels.append(channel.strip().lower())
        if 'default' not in valid_channels:
            valid_channels.append('default')
        self.channels = valid_channels
        return self.channels

    def update(self, data: dict={}, with_actions: bool=False, jpr_api_key: str='', with_healthcheck: bool=False) -> Agent:
        agent_node = super().update(data=data)
        if with_actions:
            actions = data.get('actions')
            if not isinstance(actions, list):
                return agent_node
            if len(actions) == 0:
                return agent_node
            if not jpr_api_key:
                jpr_api_key = self.jpr_api_key
            self.get_actions().install_actions(agent_id=self.id, action_list=data['actions'], jpr_api_key=jpr_api_key)
        if with_healthcheck:
            healthcheck_report = self.healthcheck()
            if healthcheck_report['status'] == 200:
                self.logger.info(healthcheck_report['message'])
            else:
                for label, detail in healthcheck_report['trace'].items():
                    if detail.get('status') == False:
                        self.logger.error(f"[{label}] {detail.get('message')}")
                    elif detail.get('status') == True and detail.get('severity') == 'warning':
                        self.logger.warning(f"[{label}] {detail.get('message')}")
        self.validate_channels()
        if agent_node:
            self.dump_descriptor()
        return agent_node

    def is_logging(self) -> bool:
        return self.agent_logging

    def set_logging(self, agent_logging: bool) -> None:
        self.agent_logging = agent_logging

    def get_descriptor(self, as_yaml: bool=False, clean: bool=False) -> Union[str, dict]:
        try:
            agent_data = {}
            agent_actions = JacList([])
            agent_ignore_keys = JacList([])
            action_ignore_keys = JacList(['_package'])
            if clean:
                agent_ignore_keys = self.protected_attrs + JacList(['meta'])
                action_ignore_keys = JacList(['id', 'description', '_package', 'weight', 'api_key', 'secret_key', 'token', 'host', 'port', 'protocol', 'api_key_name', 'connection_timeout', 'collection_name'])
            agent_data = self.export(agent_ignore_keys, clean)
            agent_actions = self.spawn(_export_actions(action_ignore_keys, clean)).action_nodes
            agent_data = {**agent_data, **{'actions': agent_actions or JacList([])}}
            if as_yaml:
                return Utils.yaml_dumps(agent_data)
            return agent_data
        except Exception as e:
            self.logger.error(f'an exception occurred, {traceback.format_exc()}')
        return {}

    def get_daf_info(self) -> dict:
        daf_info = {'package': {'name': self.meta.get('namespace', f'default/{Utils.to_snake_case(self.name)}'), 'author': self.meta.get('author', ''), 'version': self.meta.get('version', '0.0.1'), 'meta': {'title': self.name, 'description': self.description, 'type': 'daf'}}}
        if (dependencies := self.meta.get('dependencies')):
            daf_info['package']['dependencies'] = dependencies
        return daf_info

    def dump_descriptor(self) -> None:
        agent_data = self.get_descriptor(clean=False)
        if (yaml_output := Utils.yaml_dumps(agent_data)):
            self.save_file(self.descriptor, yaml_output.encode('utf-8'))
        else:
            self.logger.error('Unable to dump agent descriptor to file')

    def healthcheck(self) -> Union[bool, dict]:
        trace = {}
        if not self.published:
            trace['published'] = {'status': False, 'message': 'Agent must be published to interact with it.', 'severity': 'error'}
        if len(self.jpr_api_key.strip()) == 0:
            trace['jpr_api_key'] = {'status': True, 'message': 'JPR API key not set. Your agent will not be able to access private JIVAS package repo items.', 'severity': 'warning'}
        if isinstance(self.name, str) and len(self.name.strip()) == 0:
            trace['name'] = {'status': False, 'message': 'Agent name must be a non-empty string with no leading or trailing spaces.', 'severity': 'error'}
        if isinstance(self.description, str) and len(self.description.strip()) == 0:
            trace['description'] = {'status': False, 'message': 'Agent description must be a non-empty string with no leading or trailing spaces.', 'severity': 'error'}
        if len(self.descriptor.strip()) == 0:
            trace['descriptor'] = {'status': False, 'message': 'Agent descriptor path not set.', 'severity': 'error'}
        if not isinstance(self.message_limit, int) or self.message_limit <= 0:
            trace['message_limit'] = {'status': False, 'message': 'Message limit must be a positive integer.', 'severity': 'error'}
        if not isinstance(self.flood_block_time, int) or self.flood_block_time <= 0:
            trace['flood_block_time'] = {'status': False, 'message': 'Flood block time must be a positive integer representing seconds.', 'severity': 'error'}
        if not isinstance(self.window_time, int) or self.window_time <= 0:
            trace['window_time'] = {'status': False, 'message': 'Window time must be a positive integer representing seconds.', 'severity': 'error'}
        if not isinstance(self.flood_threshold, int) or self.flood_threshold <= 0:
            trace['flood_threshold'] = {'status': False, 'message': 'Flood threshold must be a positive integer.', 'severity': 'error'}
        if not isinstance(self.frame_size, int) or self.frame_size <= -1:
            trace['frame_size'] = {'status': False, 'message': 'Frame size must be 0 or a positive integer.', 'severity': 'error'}
        return trace

    def get_healthcheck_report(self) -> dict:
        result = {'status': 501, 'message': 'Agent healthcheck was unable to execute.', 'trace': {}}
        error_warning_trace = {}
        trace = self.healthcheck()
        trace.update(self.spawn(_healthcheck_actions()).trace)
        for key, value in trace.items():
            if isinstance(value, dict):
                if value.get('status') == False:
                    result['status'] = 503
                    error_warning_trace[key] = {'status': False, 'message': value.get('message', f'Agent healthcheck failed on {key}. Inspect configuration and try again.'), 'severity': value.get('severity')}
                elif value.get('status') == True and value.get('severity') == 'warning':
                    error_warning_trace[key] = {'status': True, 'message': value.get('message', f'Agent healthcheck warning on {key}.'), 'severity': value.get('severity')}
            elif value == False:
                result['status'] = 503
                error_warning_trace[key] = {'status': False, 'message': f'Agent healthcheck failed on {key}. Inspect configuration and try again.', 'severity': 'error'}
        if result['status'] == 501:
            result['status'] = 200
            result['message'] = 'Agent healthcheck passed.'
            result['trace'] = error_warning_trace
        elif result['status'] == 503:
            result['message'] = 'Agent healthcheck failed.'
            result['trace'] = error_warning_trace
        self.healthcheck_status = result['status']
        return result

    def get_file(self, path: str) -> bytes | None:
        return file_interface.get_file(f'{self.id}/{path}')

    def save_file(self, path: str, content: bytes) -> bool:
        return file_interface.save_file(f'{self.id}/{path}', content)

    def delete_file(self, path: str) -> bool:
        return file_interface.delete_file(f'{self.id}/{path}')

    def get_file_url(self, path: str) -> str | None:
        return file_interface.get_file_url(f'{self.id}/{path}')

    def get_short_file_url(self, path: str, with_filename: bool=False) -> str | None:
        collection = NodeAnchor.Collection.get_collection('url_proxies')
        if (url := file_interface.get_file_url(f'{self.id}/{path}')):
            if (url_proxy := collection.insert_one({'agent_id': self.id, 'path': f'{self.id}/{path}'})):
                base_url = f"{os.environ.get('JIVAS_FILES_URL', 'http://localhost:9000')}/f"
                if not with_filename:
                    return f'{base_url}/{url_proxy.inserted_id}'
                path_segments = path.split('/')
                filename = path_segments[-1]
                return f'{base_url}/{url_proxy.inserted_id}/{filename}'
        return None

class _healthcheck_actions(Walker):
    trace: dict = field(gen=lambda: {})

    class __specs__(Obj):
        private: static[bool] = True

    @with_entry
    def on_agent(self, here: Agent) -> None:
        self.visit(here.refs().filter(Actions, None))

    @with_entry
    def on_actions(self, here: Actions) -> None:
        self.visit(here.refs().filter(Action, None).filter(None, lambda item: item.enabled == True))

    @with_entry
    def on_action(self, here: Action) -> None:
        self.trace[here.label] = here.healthcheck()

class _export_actions(Walker):
    ignore_keys: list = field(gen=lambda: JacList(['_package']))
    action_nodes: list = field(gen=lambda: JacList([]))
    node_index: dict = field(gen=lambda: {})
    clean: bool = field(False)

    class __specs__(Obj):
        private: static[bool] = True

    @with_entry
    def on_agent(self, here: Agent) -> None:
        self.visit(here.refs().filter(Actions, None))

    @with_entry
    def on_actions(self, here: Actions) -> None:
        self.visit(here.refs().filter(Action, None))

    @with_entry
    def on_action(self, here: Action) -> None:
        if here.label != 'ExitInteractAction':
            children = JacList([])
            if isinstance(here, InteractAction):
                child_nodes = here.get_children()
                for child in child_nodes:
                    children.append(child.id)
            self.node_index.update({here.id: {'action': here._package['name'], 'context': here.export(self.ignore_keys, self.clean), 'children': children}})
            self.ignore(here)
        self.visit(here.refs().filter(Action, None))

    @with_exit
    def on_exit(self, here) -> None:
        if self.node_index:
            node_keys = list(self.node_index.keys())
            node_keys.reverse()
            for key in node_keys:
                resolved_nodes = JacList([])
                for child_id in self.node_index[key]['children']:
                    resolved_nodes.append(self.node_index[child_id])
                    self.node_index.pop(child_id)
                if resolved_nodes:
                    self.node_index[key]['children'] = resolved_nodes
                else:
                    self.node_index[key].pop('children')
            self.action_nodes = list(self.node_index.values())


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    from typing import Optional
else:
    Optional, = jac_import('typing', 'py', items={'Optional': None})
if typing.TYPE_CHECKING:
    from jivas.agent.memory.memory_frame_walker import memory_frame_walker
else:
    memory_frame_walker, = jac_import('jivas.agent.memory.memory_frame_walker', items={'memory_frame_walker': None})
if typing.TYPE_CHECKING:
    from jivas.agent.memory.memory import Memory
else:
    Memory, = jac_import('jivas.agent.memory.memory', items={'Memory': None})

class delete_frame(memory_frame_walker, Walker):
    session_id: str = field('')

    class __specs__(Obj):
        private: static[bool] = False

    @with_entry
    def on_memory(self, here: Memory) -> None:
        if not self.session_id:
            Jac.get_context().status = 400
            Jac.report('missing session_id')
            return
        deleted_nodes = here.purge(self.session_id)
        if not deleted_nodes:
            Jac.get_context().status = 500
            Jac.report(f'unable to delete frame with session_id {self.session_id}')
            return
        Jac.report(deleted_nodes)


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    import io
else:
    io, = jac_import('io', 'py')
if typing.TYPE_CHECKING:
    import yaml
else:
    yaml, = jac_import('yaml', 'py')
if typing.TYPE_CHECKING:
    import logging
else:
    logging, = jac_import('logging', 'py')
if typing.TYPE_CHECKING:
    import traceback
else:
    traceback, = jac_import('traceback', 'py')
if typing.TYPE_CHECKING:
    from logging import Logger
else:
    Logger, = jac_import('logging', 'py', items={'Logger': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.agent import Agent
else:
    Agent, = jac_import('jivas.agent.core.agent', items={'Agent': None})
if typing.TYPE_CHECKING:
    from jivas.agent.core.agent_graph_walker import agent_graph_walker
else:
    agent_graph_walker, = jac_import('jivas.agent.core.agent_graph_walker', items={'agent_graph_walker': None})

class install_action(agent_graph_walker, Walker):
    reporting: bool = field(False)
    override_action: bool = field(True)
    package_name: str = field('')
    version: str = field('')
    jpr_api_key: str = field('')
    logger: static[Logger] = logging.getLogger(__name__)

    class __specs__(Obj):
        private: static[bool] = False

    @with_entry
    def on_agent(self, here: Agent) -> None:
        action_exists = False
        descriptor_data = here.get_descriptor()
        if not descriptor_data:
            message = 'unable to load descriptor data'
            self.logger.error(message)
            Jac.get_context().status = 500
            Jac.report(message)
            return False
        dep_action_info = here.get_actions().get_action_info(namespace_package_name=self.package_name, version=self.version, jpr_api_key=self.jpr_api_key)
        if not dep_action_info:
            message = f'unable to locate action {self.package_name} {self.version}'
            self.logger.error(message)
            Jac.get_context().status = 403
            Jac.report(message)
            return False
        try:
            action_info = {'action': dep_action_info.get('name'), 'context': {'version': dep_action_info.get('version'), 'enabled': True}}
            for action in descriptor_data['actions']:
                if action['action'] == self.package_name:
                    if self.override_action:
                        action['context'] = action_info['context']
                        here.get_actions().deregister_action(action_label=action.get('context', {}).get('label', ''))
                    else:
                        action['context']['version'] = action_info['context']['version']
                    action_exists = True
                    break
            if not action_exists:
                descriptor_data['actions'].append(action_info)
            if here.update(data=descriptor_data, with_actions=True, jpr_api_key=self.jpr_api_key):
                message = f'{self.package_name} {self.version} installed successfully'
                self.logger.info(message)
                Jac.get_context().status = 200
                Jac.report(message)
                return True
            else:
                message = f'unable to install action {self.package_name} {self.version}'
                self.logger.error(message)
                Jac.get_context().status = 500
                Jac.report(message)
                return False
        except Exception as e:
            message = f'unable to complete operation, {traceback.format_exc()}'
            self.logger.error(message)
            Jac.get_context().status = 500
            Jac.report(message)
            return False


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    from jivas.agent.action.action import Action
else:
    Action, = jac_import('jivas.agent.action.action', items={'Action': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.interact_graph_walker import interact_graph_walker
else:
    interact_graph_walker, = jac_import('jivas.agent.action.interact_graph_walker', items={'interact_graph_walker': None})

class InteractAction(Action, Node):
    anchors: list = field(gen=lambda: JacList([]))
    functions: list = field(gen=lambda: JacList([]))
    weight: int = field(0)

    @abstract
    def touch(self, visitor: interact_graph_walker) -> bool:
        pass

    @abstract
    def execute(self, visitor: interact_graph_walker) -> dict:
        pass

    def deny(self, visitor: interact_graph_walker) -> dict:
        pass

    def get_children(self) -> None:
        return self.refs().filter(Action, None)

    def get_root_action(self) -> None:
        return self.spawn(_get_root_action()).root_action

class _get_root_action(Walker):
    root_action: InteractAction = field(None)

    class __specs__(Obj):
        private: static[bool] = True

    @with_entry
    def on_action(self, here: InteractAction) -> None:
        if not self.visit(here.refs(dir=EdgeDir.IN).filter(InteractAction, None)):
            self.root_action = here


from __future__ import annotations
from jaclang import *
import typing
from enum import Enum, auto
if typing.TYPE_CHECKING:
    from jivas.agent.core.graph_node import GraphNode
else:
    GraphNode, = jac_import('jivas.agent.core.graph_node', items={'GraphNode': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.interact_action import InteractAction
else:
    InteractAction, = jac_import('jivas.agent.action.interact_action', items={'InteractAction': None})
if typing.TYPE_CHECKING:
    from jivas.agent.action.interact_graph_walker import interact_graph_walker
else:
    interact_graph_walker, = jac_import('jivas.agent.action.interact_graph_walker', items={'interact_graph_walker': None})

class ExitInteractAction(InteractAction, Node):
    label: str = field('ExitInteractAction')
    description: str = field('core exit action node for walker cleanup and return')
    weight: int = field(10000)

    def touch(self, visitor: interact_graph_walker) -> bool:
        return True

    def execute(self, visitor: interact_graph_walker) -> dict:
        return {}